#version 100

#ifdef GL_ES
#ifdef GL_FRAGMENT_PRECISION_HIGH
precision highp float;
#else
precision mediump float;
#endif
#endif

uniform sampler2D texture;
varying vec2 texCoord;

uniform float brightnessValue;
uniform float contrastValue;
uniform float saturationValue;
uniform vec3 tintHue;
uniform float tintIntensity;
uniform vec2 blurRadius;
uniform sampler2D unsharpMaskBlurredTexture;
uniform float unsharpMaskStrength;
uniform sampler2D xproTexture;
uniform float xproStrength;

uniform float vignetteStrength;
uniform vec2 vignetteCenter;
uniform vec2 vignetteScale;
uniform float vignetteMinRadius;
uniform float vignetteMaxRadius;

uniform sampler2D alphaMaskTexture;
uniform bool isMasked;
uniform int maskChannel;

uniform float opacity;

uniform float offsetTop;
uniform float offsetBottom;
uniform float offsetRight;
uniform float offsetLeft;

uniform float highlightsValue;
uniform float warmthValue;
uniform float vibranceValue;
uniform float shadowsValue;
uniform float fadeValue;
uniform float clarityValue;
uniform sampler2D clarityMaskBlurredTexture;

// Transformation matrix
// This matrix transforms the color space so that brightness is represented in
// the red color channel as per:
// https://canvadev.atlassian.net/wiki/spaces/~652209499/pages/1931510075/Highlights+and+Shadows
const mat4 M1 = mat4(0.255, 0.341, 0.105, 0.000, 0.652, -0.465, 0.319, 0.000, 0.093, 0.125, -0.424, 0.000, 0.000, 0.500, 0.500, 1.000);

// Inverse transformation matrix
const mat4 M2 = mat4(1.000, 1.000, 1.000, 0.000, 1.944, -0.748, -0.080, 0.000, 0.793, 0.000, -2.161, 0.000, -1.368, 0.374, 1.121, 1.000);

vec4 brightness(vec4 color, float value) {
  return vec4(color.rgb + value, color.a);
}

vec4 contrast(vec4 color, float value) {
  if (value > 0.0) {
    color.rgb = (color.rgb - 0.5) / (1.0 - value) + 0.5;
  } else {
    color.rgb = (color.rgb - 0.5) * (1.0 + value) + 0.5;
  }
  return color;
}

vec4 saturation(vec4 color, float value) {
  float sv = value / 100.0 + 1.0;

  vec3 luminanceWeighting = vec3(0.3086, 0.6094, 0.0820);
  float luminance = dot(luminanceWeighting, color.rgb);
  vec3 greyScale = vec3(luminance);

  return vec4(mix(greyScale, color.rgb, sv), color.a);
}

vec4 tint(vec4 color, vec3 hue, float intensity) {
  color.rgb -= vec3(
    (color.r - hue.r) * intensity,
    (color.g - hue.g) * intensity,
    (color.b - hue.b) * intensity
  );
  return color;
}

float rand(vec2 co) {
  float a = 12.9898;
  float b = 78.233;
  float c = 43758.5453;
  float dt = dot(co.xy ,vec2(a,b));
  float sn = mod(dt,3.14);
  return fract(sin(sn) * c);
}

vec4 blur(vec2 radius) {
  vec4 color = vec4(0.0);
  float total = 0.0;

  /* randomize the lookup values to hide the fixed number of samples */
  float offset = rand(gl_FragCoord.xy);

  float size = 10.0; // TODO(Theeban): this value needs to be based on texture size
  float start = -size + offset - 0.5;
  float end = size + offset - 0.5;
  for (float t = start; t <= end; t++) {
    float percent = t / size;
    float weight = 1.0 - abs(percent);
    vec2 sampleVec = texCoord + radius * percent;
    if (sampleVec.x <= 1.0 && sampleVec.y <= 1.0) {
      vec4 sampleColor = texture2D(texture, sampleVec);
      color += sampleColor * weight;
      total += weight;
    }
  }

  color /= total;

  return color;
}

vec4 unsharpMask(sampler2D originalTexture, sampler2D blurredTexture, float strength) {
  vec4 original = texture2D(originalTexture, texCoord);
  vec4 blurred = texture2D(blurredTexture, texCoord);
  return mix(blurred, original, 1.0 + strength);
}

vec4 vignette() {
  vec4 color = texture2D(texture, texCoord);
  // make sure transparent pixels don't have any color
  if (color.a == 0.0) {
    color.rgb *= 0.0;
  }
  vec2 offset = (texCoord - vignetteCenter) / vignetteScale;
  float radius = length(offset);
  float alpha  = clamp((radius - vignetteMinRadius) / (vignetteMaxRadius - vignetteMinRadius), 0.0, 1.0);
  float t = alpha * alpha * (3.0 - 2.0 * alpha);
  return mix(color, vec4(0.0,0.0,0.0,1.0), vignetteStrength * t);
}

/**
  * Lookup the textcoord of the center pixel in a 256 pixel texture.
  */
float lookupPixelCenter(float v) {
  float index = floor(clamp(v, 0.0, 1.0) * 255.0);
  float center = index + 0.5;
  return center / 256.0;
}

vec4 curves(vec4 color) {
  float redCurveValue = texture2D(xproTexture, vec2(lookupPixelCenter(color.r), 0.0)).r;
  float greenCurveValue = texture2D(xproTexture, vec2(lookupPixelCenter(color.g), 0.0)).g;
  float blueCurveValue = texture2D(xproTexture, vec2(lookupPixelCenter(color.b), 0.0)).b;
  return vec4(redCurveValue, greenCurveValue, blueCurveValue, color.a);
}

vec4 applyHighlights(vec4 color, float highlights) {
  // Apply transformation matrix
  vec4 transformedColor = M1 * color;

  // Set up LUT
  float scaleAmount = 0.8 * step(0.0, highlights) + 0.2;  // 0.2 when amount < 0, 1.0 otherwise
  float alpha = pow(transformedColor.r, 3.0);
  float newBrightness = clamp(transformedColor.r * (1.0 + highlights * scaleAmount), 0.0, 1.0);

  // Apply LUT
  transformedColor.r = mix(transformedColor.r, newBrightness, alpha);

  // Apply inverse transformation matrix
  color = M2 * transformedColor;
  color.r = clamp(color.r, 0.0, 1.0);
  return color;
}

vec4 applyShadows(vec4 color, float shadows) {
  // Apply transformation matrix
  vec4 transformedColor = M1 * color;

  // Set up LUT
  float alpha = pow(1.0 - transformedColor.r, 3.0);
  float newBrightness = clamp(transformedColor.r * (1.0 + shadows), 0.0, 1.0);

  // Apply LUT
  transformedColor.r = mix(transformedColor.r, newBrightness, alpha);

  // Apply inverse transformation matrix
  color = M2 * transformedColor;
  // TODO: Why is this only clamping red?
  color.r = clamp(color.r, 0.0, 1.0);
  return color;
}

uniform sampler2D uFadeLUT;
vec4 applyFade(vec4 color, float fade) {
  return vec4(mix(
    color.rgb,
    vec3(
      texture2D(uFadeLUT, vec2(color.r, 0.5)).r,
      texture2D(uFadeLUT, vec2(color.g, 0.5)).g,
      texture2D(uFadeLUT, vec2(color.b, 0.5)).b
    ),
    fade
  ), color.a);
}

// Vibrance
float getDistance(in vec2 gb) {
  gb = gb * 2.0 - 1.0;  // rescale to -1 to 1
  gb.y *= 1.25 - 0.5 * step(0.0, gb.y);  // scale negative values by 1.25 scale positive by 0.5
  gb.xy *= gb.xy;        // square green and blue channels
  gb.xy = clamp(gb.xy, 0.0, 1.0);
  gb *= 255.0;  // to implement the discretizing that occurs in the FE
  gb = floor(gb);
  return sqrt(clamp(gb.x + gb.y, 0.0, 255.0) / 255.0);
}

vec4 applySaturation(vec4 color, float sv) {
  // Resources:
  // https://www.w3.org/TR/SVG11/filters.html#feColorMatrixElement
  // Saturation is equivalent to this matrix operation:
  // | R' |     |0.213+0.787s  0.715-0.715s  0.072-0.072s 0  0 |   | R |
  // | G' |     |0.213-0.213s  0.715+0.285s  0.072-0.072s 0  0 |   | G |
  // | B' |  =  |0.213-0.213s  0.715-0.715s  0.072+0.928s 0  0 | * | B |
  // | A' |     |           0            0             0  1  0 |   | A |
  // | 1  |     |           0            0             0  0  1 |   | 1 |
  vec3 luminance = vec3(0.2125, 0.7154, 0.0721);
  vec3 v = vec3(1.0 - sv) * luminance;
  mat3 transform = mat3(vec3(v.x) + vec3(sv, 0, 0), vec3(v.y) + vec3(0, sv, 0), vec3(v.z) + vec3(0, 0, sv));
  return vec4(clamp(transform * color.rgb, 0.0, 1.0), color.a);
}

float saturationMask(in float x) { return pow(1.0 - x, 3.0); }

// Implements vibrance filter as described in
// https://canvadev.atlassian.net/wiki/spaces/~652209499/pages/1964279793/Vibrance
vec4 applyVibrance(vec4 color, float vibrance) {
  // Transformation matrix
  // This matrix transforms the RGB cube so that the grayscale colors are facing
  // the x axis
  const mat3 M1 = mat3(0.334, 0, -0.5, 0.333, 0.434, 0.25, 0.333, -0.433, 0.251);

  // Apply a 5D transformation matrix in 2 steps
  vec3 transformedColor = M1 * color.rgb;
  transformedColor.gb += 0.5;

  // Saturation is given by distance of transformed color from the axis
  float distance = getDistance(transformedColor.gb);

  // Create a mask that is weighted towards desaturated colors
  float satMask = saturationMask(clamp(distance, 0.0, 1.0));

  // Apply the mask using "in" composite operation
  vec4 maskedInput = satMask * color;

  // Increase saturation of resulting image
  vec4 satResult = applySaturation(maskedInput, 1.0 + vibrance);

  // Combine result with original image using "atop" composite operation
  return color.a * satResult + (1.0 - satResult.a) * color;
}

// Warmth
vec4 applyWarmth(vec4 color, float warmth) {
  return vec4(color.r + warmth, color.g, color.b - warmth, color.a);
}

// Clarity
// converts from rgb to luminance, using operation from here
// https://github.com/WebKit/WebKit/blob/88278b55563e5ccdc0b3419c6c391c3becc19e40/Source/WebCore/platform/graphics/filters/FEColorMatrix.cpp
float getLuminance(vec3 rgb) {
  vec3 luminance = vec3(0.2125, 0.7154, 0.0721);
  return dot(rgb, luminance);
}

// linearly interpolates values in input range to those in specified output range
float mapBetween(float value, float inMin, float inMax, float outMin, float outMax) {
  return outMin + (outMax - outMin) * (value - inMin) / (inMax - inMin);
}

// linearly interpolates luminance values between a number of ranges:
// luminance in range 0 to 1/7 mapped to 0 to 0.75
// luminance in range 1/7 to 2/7 mapped to 0.75 to 1
// luminance in range 2/7 to 5/7 mapped to 1
// luminance in range 5/7 to 6/7 mapped to 0.75 to 1
// luminance in range 6/7 to 1/7 mapped to 0 to 0.75
float midtonesMask(float luminance) {
  float range1 = 1.0 - step(0.1429, luminance);
  float range2 = step(0.1429, luminance) - step(0.2857, luminance);
  float range3 = step(0.2857, luminance) - step(0.7143, luminance);
  float range4 = step(0.7143, luminance) - step(0.8571, luminance);
  float range5 = step(0.8571, luminance);

  float luminanceMapped = mapBetween(luminance, 0.0, 0.1429, 0.0, 0.75) * range1;
  luminanceMapped += mapBetween(luminance, 0.1429, 0.2857, 0.75, 1.0) * range2;
  luminanceMapped += mapBetween(luminance, 0.2857, 0.7143, 1.0, 1.0) * range3;
  luminanceMapped += mapBetween(luminance, 0.7143, 0.8571, 1.0, 0.75) * range4;
  luminanceMapped += mapBetween(luminance, 0.8571, 1.0, 0.75, 0.0) * range5;

  return luminanceMapped;
}

vec4 clarity(vec4 color, sampler2D blurredTexture, float clarity) {
  float luminance = getLuminance(color.rgb);
  vec4 midtones = vec4(color.rgb, color.a * midtonesMask(luminance));
  vec4 blurred = texture2D(blurredTexture, texCoord);
  midtones.rgb *= midtones.a;
  color.rgb *= color.a;
  blurred.a = 1.0;

  // This is equivalent to the frontend arithmetic composite e.g.:
  // k2 = blurredColor * -strength
  // k3 = color * (1.0 + strength)
  // K2 + k3
  // The output of which is clamped between 0 and 1
  vec4 clarityClip = mix(blurred, midtones, 1.0 + clarity);
  clarityClip = clamp(clarityClip, 0.0, 1.0);

  // Output composited with input via "atop" operator
  color = color.a * clarityClip + (1.0 - clarityClip.a) * color;
  color.rgb /= color.a;

  return color;
}

void main() {
  vec4 color = texture2D(texture, texCoord);
  if (brightnessValue != 0.0) {
    color = brightness(color, brightnessValue);
  }
  if (contrastValue != 0.0) {
    color = contrast(color, contrastValue);
  }
  if (saturationValue != 0.0) {
    color = saturation(color, saturationValue);
  }
  if (tintHue != vec3(0.0, 0.0, 0.0) && tintIntensity != 0.0) {
    color = tint(color, tintHue, tintIntensity);
  }
  if (highlightsValue != 0.0) {
    color = applyHighlights(color, highlightsValue);
  }
  if (warmthValue != 0.0) {
    color = applyWarmth(color, warmthValue);
  }
  if (vibranceValue != 0.0) {
    color = applyVibrance(color, vibranceValue);
  }
  if (shadowsValue != 0.0) {
    color = applyShadows(color, shadowsValue);
  }
  if (fadeValue != 0.0) {
    color = applyFade(color, fadeValue);
  }
  if (xproStrength != 0.0) {
    color = curves(color);
  }
  if (blurRadius != vec2(0.0, 0.0)) {
    color = blur(blurRadius);
  }
  if (unsharpMaskStrength != 0.0) {
    color = unsharpMask(texture, unsharpMaskBlurredTexture, unsharpMaskStrength);
  }
  if (clarityValue != 0.0) {
    color = clarity(color, clarityMaskBlurredTexture, clarityValue);
  }
  if (vignetteStrength != 0.0) {
    color = vignette();
  }

  float opacityFactor;
  if (isMasked) {
    vec4 mask = texture2D(alphaMaskTexture, texCoord);
    opacityFactor = opacity * mask[maskChannel];
  } else {
    opacityFactor = opacity;
  }
  color.a *= opacityFactor;

  //make sure outbound of texture pixels is transparent
  bool clearColor = texCoord.y > offsetTop || texCoord.y < offsetBottom || texCoord.x > offsetRight || texCoord.x < offsetLeft;
  
  gl_FragColor = color * vec4(!clearColor);
}
