// from https://github.com/gl-transitions/gl-transitions

#version 100

#ifdef GL_ES
#ifdef GL_FRAGMENT_PRECISION_HIGH
precision highp float;
#else
precision mediump float;
#endif
#endif

uniform sampler2D texture;
varying vec2 texCoord;

uniform sampler2D fromTex;
uniform sampler2D toTex;
uniform float progress;
uniform vec2 direction;
uniform bool opening;
uniform float ratio;
uniform bool clockwise;
uniform vec2 origin;

uniform bool wipe;
uniform bool wipe_line;
uniform bool wipe_circle;
uniform bool slide;
uniform bool dissolve;
uniform bool chop;
uniform bool stack;
uniform bool color_wipe;
uniform bool flow;

const float smoothness = 0.2;
const vec2 center = vec2(0.5, 0.5);
const float PI = 3.14159265359;

vec4 getFromColor(vec2 uv){ return texture2D(fromTex, uv); }
vec4 getToColor(vec2 uv){ return texture2D(toTex, uv); }

vec4 wipeTransition(vec2 uv) {
    vec2 v = normalize(direction);
    v /= abs(v.x)+abs(v.y);
    float d = v.x * center.x + v.y * center.y;
    float m =
    (1.0 - step(progress, 0.0)) *// there is something wrong with our formula that makes m not equals 0.0 with progress is 0.0
    (1.0 - smoothstep(-smoothness, 0.0, v.x * uv.x + v.y * uv.y - (d - 0.5 + progress * (1.0 + smoothness))));
    return mix(getFromColor(uv), getToColor(uv), m);
}

vec4 wipeLineTransition(vec2 uv) {
    vec2 v = normalize(direction);
    v /= abs(v.x)+abs(v.y);
    float d = v.x * center.x + v.y * center.y;
    float m =
    (1.0-step(progress, 0.0)) *// there is something wrong with our formula that makes m not equals 0.0 with progress is 0.0
    (1.0 - step(0.0, v.x * uv.x + v.y * uv.y - (d-0.5+progress)));
    return mix(getFromColor(uv), getToColor(uv), m);
}

vec4 wipeCircleTransition(vec2 uv) {
    float frame_width = ratio;
    float frame_height = 1.;
    float circle_radius = sqrt(pow(frame_width / 2., 2.) + pow(frame_height / 2., 2.));

    float ellipse_width = circle_radius / ratio;
    float ellipse_height = circle_radius;

    float v = pow(uv.x - 0.5, 2.) / pow(ellipse_width, 2.) + pow((uv.y - 0.5), 2.) / pow(ellipse_height, 2.);
    bool shouldUseDestinationColor = opening ? v <= progress : v >= (1. - progress);
    return mix(getFromColor(uv), getToColor(uv), shouldUseDestinationColor ? 1. : 0.);
}

vec4 slideTransition(vec2 uv) {
    vec2 p = uv + progress * sign(direction);
    vec2 f = fract(p);
    return mix(
    getToColor(f),
    getFromColor(f),
    step(0.0, p.y) * step(p.y, 1.0) * step(0.0, p.x) * step(p.x, 1.0)
    );
}

vec4 dissolveTransition(vec2 uv) {
    return mix(getFromColor(uv), getToColor(uv), progress);
}

vec4 chopTransition(vec2 uv) {
    vec2 translatedCoordinate = uv - origin;
    translatedCoordinate.x *= ratio;

    float direction = clockwise ? 1.0 : -1.0;
    float quarter = PI / 2.0;
    float angle = direction * (3.0 * quarter + quarter * progress);

    float c = cos(angle);
    float s = sin(angle);

    vec2 rotated = vec2(
    translatedCoordinate.x * c - translatedCoordinate.y * s,
    translatedCoordinate.y * c + translatedCoordinate.x * s
    );

    rotated.x /= ratio;
    vec2 rotatedUv = rotated + origin;

    if (rotatedUv.x < 0.0 || rotatedUv.x > 1.0 ||
    rotatedUv.y < 0.0 || rotatedUv.y > 1.0) {
        return getFromColor(uv);
    } else {
        return getToColor(rotatedUv);
    }
}

vec4 stackTransition(vec2 uv) {
    vec2 p = uv + progress * sign(direction);
    vec2 f = fract(p);
    return mix(
    getToColor(f),
    getFromColor(uv),
    step(0.0, p.y) * step(p.y, 1.0) * step(0.0, p.x) * step(p.x, 1.0)
    );
}

vec4 colorWipeTransition(vec2 uv) {
    if (progress < 0.5) {
        return getFromColor(uv);
    } else {
        return getToColor(uv);
    }
}

vec4 flowTransition(vec2 uv) {
    vec4 toColor = getToColor(uv);
    float toColorA = toColor.a;
    toColor.a = 1.0;
    return mix(getFromColor(uv), toColor, toColorA);
}

void main() {
    gl_FragColor =
    wipeTransition(texCoord) * vec4(wipe) +
    wipeLineTransition(texCoord) * vec4(wipe_line) +
    wipeCircleTransition(texCoord) * vec4(wipe_circle) +
    slideTransition(texCoord) * vec4(slide) +
    dissolveTransition(texCoord) * vec4(dissolve) +
    chopTransition(texCoord) * vec4(chop) +
    stackTransition(texCoord) * vec4(stack) +
    colorWipeTransition(texCoord) * vec4(color_wipe) +
    flowTransition(texCoord) * vec4(flow);
}
